"""
Test for DFT Utilities
"""
try:
    import dqc
    from dqc.system.mol import Mol
    from dqc.qccalc.ks import KS
    from deepchem.utils.dftutils import KSCalc, hashstr
    import torch
except ModuleNotFoundError:
    pass
import pytest


@pytest.mark.dqc
def test_dftutils():
    system = {
        'type': 'mol',
        'kwargs': {
            'moldesc': 'H 0.86625 0 0; F -0.86625 0 0',
            'basis': '6-311++G(3df,3pd)'
        }
    }
    atomzs, atomposs = dqc.parse_moldesc(system["kwargs"]["moldesc"])
    mol = Mol(**system["kwargs"])
    qc = KS(mol, xc='lda_x').run()
    qcs = KSCalc(qc)
    a = qcs.energy()
    b = torch.tensor(-99.1360, dtype=torch.float64)
    assert torch.allclose(a, b)


@pytest.mark.dqc
def test_str():
    s = "hydrogen fluoride"
    s = hashstr(s)
    s1 = "df4e3775493a2e784618edaf9e96b7ecb6ce2b4cd022e8619588d55009872bb2"
    assert s == s1



"""
DGL-based GAT for graph property prediction.
"""
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel
from typing import Optional


class GAT(nn.Module):
    """Model for Graph Property Prediction Based on Graph Attention Networks (GAT).

    This model proceeds as follows:

    * Update node representations in graphs with a variant of GAT
    * For each graph, compute its representation by 1) a weighted sum of the node
        representations in the graph, where the weights are computed by applying a
        gating function to the node representations 2) a max pooling of the node
        representations 3) concatenating the output of 1) and 2)
    * Perform the final prediction using an MLP

    Examples
    --------

    >>> import deepchem as dc
    >>> import dgl
    >>> from deepchem.models import GAT
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> featurizer = dc.feat.MolGraphConvFeaturizer()
    >>> graphs = featurizer.featurize(smiles)
    >>> print(type(graphs[0]))
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> dgl_graphs = [graphs[i].to_dgl_graph(self_loop=True) for i in range(len(graphs))]
    >>> # Batch two graphs into a graph of two connected components
    >>> batch_dgl_graph = dgl.batch(dgl_graphs)
    >>> model = GAT(n_tasks=1, mode='regression')
    >>> preds = model(batch_dgl_graph)
    >>> print(type(preds))
    <class 'torch.Tensor'>
    >>> preds.shape == (2, 1)
    True

    References
    ----------
    .. [1] Petar Veličković, Guillem Cucurull, Arantxa Casanova, Adriana Romero, Pietro Liò,
        and Yoshua Bengio. "Graph Attention Networks." ICLR 2018.

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.
    """

    def __init__(self,
                 n_tasks: int,
                 graph_attention_layers: Optional[list] = None,
                 n_attention_heads: int = 8,
                 agg_modes: Optional[list] = None,
                 activation=F.elu,
                 residual: bool = True,
                 dropout: float = 0.,
                 alpha: float = 0.2,
                 predictor_hidden_feats: int = 128,
                 predictor_dropout: float = 0.,
                 mode: str = 'regression',
                 number_atom_features: int = 30,
                 n_classes: int = 2,
                 nfeat_name: str = 'x'):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        graph_attention_layers: list of int
            Width of channels per attention head for GAT layers. graph_attention_layers[i]
            gives the width of channel for each attention head for the i-th GAT layer. If
            both ``graph_attention_layers`` and ``agg_modes`` are specified, they should have
            equal length. If not specified, the default value will be [8, 8].
        n_attention_heads: int
            Number of attention heads in each GAT layer.
        agg_modes: list of str
            The way to aggregate multi-head attention results for each GAT layer, which can be
            either 'flatten' for concatenating all-head results or 'mean' for averaging all-head
            results. ``agg_modes[i]`` gives the way to aggregate multi-head attention results for
            the i-th GAT layer. If both ``graph_attention_layers`` and ``agg_modes`` are
            specified, they should have equal length. If not specified, the model will flatten
            multi-head results for intermediate GAT layers and compute mean of multi-head results
            for the last GAT layer.
        activation: activation function or None
            The activation function to apply to the aggregated multi-head results for each GAT
            layer. If not specified, the default value will be ELU.
        residual: bool
            Whether to add a residual connection within each GAT layer. Default to True.
        dropout: float
            The dropout probability within each GAT layer. Default to 0.
        alpha: float
            A hyperparameter in LeakyReLU, which is the slope for negative values. Default to 0.2.
        predictor_hidden_feats: int
            The size for hidden representations in the output MLP predictor. Default to 128.
        predictor_dropout: float
            The dropout probability in the output MLP predictor. Default to 0.
        mode: str
            The model type, 'classification' or 'regression'. Default to 'regression'.
        number_atom_features: int
            The length of the initial atom feature vectors. Default to 30.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 2.
        nfeat_name: str
            For an input graph ``g``, the model assumes that it stores node features in
            ``g.ndata[nfeat_name]`` and will retrieve input node features from that.
            Default to 'x'.
        """
        try:
            import dgl  # noqa: F401
        except:
            raise ImportError('This class requires dgl.')
        try:
            import dgllife  # noqa: F401
        except:
            raise ImportError('This class requires dgllife.')

        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        super(GAT, self).__init__()

        self.n_tasks = n_tasks
        self.mode = mode
        self.n_classes = n_classes
        self.nfeat_name = nfeat_name
        if mode == 'classification':
            out_size = n_tasks * n_classes
        else:
            out_size = n_tasks

        from dgllife.model import GATPredictor as DGLGATPredictor

        if isinstance(graph_attention_layers, list) and isinstance(
                agg_modes, list):
            assert len(graph_attention_layers) == len(agg_modes), \
              'Expect graph_attention_layers and agg_modes to have equal length, ' \
              'got {:d} and {:d}'.format(len(graph_attention_layers), len(agg_modes))

        # Decide first number of GAT layers
        if graph_attention_layers is not None:
            num_gnn_layers = len(graph_attention_layers)
        elif agg_modes is not None:
            num_gnn_layers = len(agg_modes)
        else:
            num_gnn_layers = 2

        if graph_attention_layers is None:
            graph_attention_layers = [8] * num_gnn_layers
        if agg_modes is None:
            agg_modes = ['flatten' for _ in range(num_gnn_layers - 1)]
            agg_modes.append('mean')

        if activation is not None:
            activation = [activation] * num_gnn_layers

        self.model = DGLGATPredictor(
            in_feats=number_atom_features,
            hidden_feats=graph_attention_layers,
            num_heads=[n_attention_heads] * num_gnn_layers,
            feat_drops=[dropout] * num_gnn_layers,
            attn_drops=[dropout] * num_gnn_layers,
            alphas=[alpha] * num_gnn_layers,
            residuals=[residual] * num_gnn_layers,
            agg_modes=agg_modes,
            activations=activation,
            n_tasks=out_size,
            predictor_hidden_feats=predictor_hidden_feats,
            predictor_dropout=predictor_dropout)

    def forward(self, g):
        """Predict graph labels

        Parameters
        ----------
        g: DGLGraph
            A DGLGraph for a batch of graphs. It stores the node features in
            ``dgl_graph.ndata[self.nfeat_name]``.

        Returns
        -------
        torch.Tensor
            The model output.

        * When self.mode = 'regression',
            its shape will be ``(dgl_graph.batch_size, self.n_tasks)``.
        * When self.mode = 'classification', the output consists of probabilities
            for classes. Its shape will be
            ``(dgl_graph.batch_size, self.n_tasks, self.n_classes)`` if self.n_tasks > 1;
            its shape will be ``(dgl_graph.batch_size, self.n_classes)`` if self.n_tasks is 1.

        torch.Tensor, optional
            This is only returned when self.mode = 'classification', the output consists of the
            logits for classes before softmax.
        """
        node_feats = g.ndata[self.nfeat_name]
        out = self.model(g, node_feats)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = out.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = out.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        else:
            return out


class GATModel(TorchModel):
    """Model for Graph Property Prediction Based on Graph Attention Networks (GAT).

    This model proceeds as follows:

    * Update node representations in graphs with a variant of GAT
    * For each graph, compute its representation by 1) a weighted sum of the node
        representations in the graph, where the weights are computed by applying a
        gating function to the node representations 2) a max pooling of the node
        representations 3) concatenating the output of 1) and 2)
    * Perform the final prediction using an MLP

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models import GATModel
    >>> # preparing dataset
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> labels = [0., 1.]
    >>> featurizer = dc.feat.MolGraphConvFeaturizer()
    >>> X = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
    >>> # training model
    >>> model = GATModel(mode='classification', n_tasks=1,
    ...                  batch_size=16, learning_rate=0.001)
    >>> loss = model.fit(dataset, nb_epoch=5)

    References
    ----------
    .. [1] Petar Veličković, Guillem Cucurull, Arantxa Casanova, Adriana Romero, Pietro Liò,
        and Yoshua Bengio. "Graph Attention Networks." ICLR 2018.

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.
    """

    def __init__(self,
                 n_tasks: int,
                 graph_attention_layers: Optional[list] = None,
                 n_attention_heads: int = 8,
                 agg_modes: Optional[list] = None,
                 activation=F.elu,
                 residual: bool = True,
                 dropout: float = 0.,
                 alpha: float = 0.2,
                 predictor_hidden_feats: int = 128,
                 predictor_dropout: float = 0.,
                 mode: str = 'regression',
                 number_atom_features: int = 30,
                 n_classes: int = 2,
                 self_loop: bool = True,
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        graph_attention_layers: list of int
            Width of channels per attention head for GAT layers. graph_attention_layers[i]
            gives the width of channel for each attention head for the i-th GAT layer. If
            both ``graph_attention_layers`` and ``agg_modes`` are specified, they should have
            equal length. If not specified, the default value will be [8, 8].
        n_attention_heads: int
            Number of attention heads in each GAT layer.
        agg_modes: list of str
            The way to aggregate multi-head attention results for each GAT layer, which can be
            either 'flatten' for concatenating all-head results or 'mean' for averaging all-head
            results. ``agg_modes[i]`` gives the way to aggregate multi-head attention results for
            the i-th GAT layer. If both ``graph_attention_layers`` and ``agg_modes`` are
            specified, they should have equal length. If not specified, the model will flatten
            multi-head results for intermediate GAT layers and compute mean of multi-head results
            for the last GAT layer.
        activation: activation function or None
            The activation function to apply to the aggregated multi-head results for each GAT
            layer. If not specified, the default value will be ELU.
        residual: bool
            Whether to add a residual connection within each GAT layer. Default to True.
        dropout: float
            The dropout probability within each GAT layer. Default to 0.
        alpha: float
            A hyperparameter in LeakyReLU, which is the slope for negative values. Default to 0.2.
        predictor_hidden_feats: int
            The size for hidden representations in the output MLP predictor. Default to 128.
        predictor_dropout: float
            The dropout probability in the output MLP predictor. Default to 0.
        mode: str
            The model type, 'classification' or 'regression'. Default to 'regression'.
        number_atom_features: int
            The length of the initial atom feature vectors. Default to 30.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 2.
        self_loop: bool
            Whether to add self loops for the nodes, i.e. edges from nodes to themselves.
            When input graphs have isolated nodes, self loops allow preserving the original feature
            of them in message passing. Default to True.
        kwargs
            This can include any keyword argument of TorchModel.
        """
        model = GAT(n_tasks=n_tasks,
                    graph_attention_layers=graph_attention_layers,
                    n_attention_heads=n_attention_heads,
                    agg_modes=agg_modes,
                    activation=activation,
                    residual=residual,
                    dropout=dropout,
                    alpha=alpha,
                    predictor_hidden_feats=predictor_hidden_feats,
                    predictor_dropout=predictor_dropout,
                    mode=mode,
                    number_atom_features=number_atom_features,
                    n_classes=n_classes)
        if mode == 'regression':
            loss: Loss = L2Loss()
            output_types = ['prediction']
        else:
            loss = SparseSoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
        super(GATModel, self).__init__(model,
                                       loss=loss,
                                       output_types=output_types,
                                       **kwargs)

        self._self_loop = self_loop

    def _prepare_batch(self, batch):
        """Create batch data for GAT.

        Parameters
        ----------
        batch: tuple
            The tuple is ``(inputs, labels, weights)``.

        Returns
        -------
        inputs: DGLGraph
            DGLGraph for a batch of graphs.
        labels: list of torch.Tensor or None
            The graph labels.
        weights: list of torch.Tensor or None
            The weights for each sample or sample/task pair converted to torch.Tensor.
        """
        try:
            import dgl
        except:
            raise ImportError('This class requires dgl.')

        inputs, labels, weights = batch
        dgl_graphs = [
            graph.to_dgl_graph(self_loop=self._self_loop) for graph in inputs[0]
        ]
        inputs = dgl.batch(dgl_graphs).to(self.device)
        _, labels, weights = super(GATModel, self)._prepare_batch(
            ([], labels, weights))
        return inputs, labels, weights



"""
Computes putative binding pockets on protein.
"""
import logging
import numpy as np
from typing import Any, List, Optional, Tuple

from deepchem.models import Model
from deepchem.utils.rdkit_utils import load_molecule
from deepchem.utils.coordinate_box_utils import CoordinateBox, get_face_boxes, merge_overlapping_boxes
from deepchem.utils.fragment_utils import get_contact_atom_indices

logger = logging.getLogger(__name__)


def extract_active_site(
        protein_file: str,
        ligand_file: str,
        cutoff: float = 4.0) -> Tuple[CoordinateBox, np.ndarray]:
    """Extracts a box for the active site.

    Parameters
    ----------
    protein_file: str
        Location of protein PDB
    ligand_file: str
        Location of ligand input file
    cutoff: float, optional (default 4.0)
        The distance in angstroms from the protein pocket to
        consider for featurization.

    Returns
    -------
    Tuple[CoordinateBox, np.ndarray]
        A tuple of `(CoordinateBox, np.ndarray)` where the second entry is
        of shape `(N, 3)` with `N` the number of atoms in the active site.
    """
    protein = load_molecule(protein_file, add_hydrogens=False)
    ligand = load_molecule(ligand_file, add_hydrogens=True, calc_charges=True)
    protein_contacts, ligand_contacts = get_contact_atom_indices(
        [protein, ligand], cutoff=cutoff)
    protein_coords = protein[0]
    pocket_coords = protein_coords[protein_contacts]

    x_min = int(np.floor(np.amin(pocket_coords[:, 0])))
    x_max = int(np.ceil(np.amax(pocket_coords[:, 0])))
    y_min = int(np.floor(np.amin(pocket_coords[:, 1])))
    y_max = int(np.ceil(np.amax(pocket_coords[:, 1])))
    z_min = int(np.floor(np.amin(pocket_coords[:, 2])))
    z_max = int(np.ceil(np.amax(pocket_coords[:, 2])))
    box = CoordinateBox((x_min, x_max), (y_min, y_max), (z_min, z_max))
    return box, pocket_coords


class BindingPocketFinder(object):
    """Abstract superclass for binding pocket detectors

    Many times when working with a new protein or other macromolecule,
    it's not clear what zones of the macromolecule may be good targets
    for potential ligands or other molecules to interact with. This
    abstract class provides a template for child classes that
    algorithmically locate potential binding pockets that are good
    potential interaction sites.

    Note that potential interactions sites can be found by many
    different methods, and that this abstract class doesn't specify the
    technique to be used.
    """

    def find_pockets(self, molecule: Any):
        """Finds potential binding pockets in proteins.

        Parameters
        ----------
        molecule: object
            Some representation of a molecule.
        """
        raise NotImplementedError


class ConvexHullPocketFinder(BindingPocketFinder):
    """Implementation that uses convex hull of protein to find pockets.

    Based on https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4112621/pdf/1472-6807-14-18.pdf
    """

    def __init__(self, scoring_model: Optional[Model] = None, pad: float = 5.0):
        """Initialize the pocket finder.

        Parameters
        ----------
        scoring_model: Model, optional (default None)
            If specified, use this model to prune pockets.
        pad: float, optional (default 5.0)
            The number of angstroms to pad around a binding pocket's atoms
            to get a binding pocket box.
        """
        self.scoring_model = scoring_model
        self.pad = pad

    def find_all_pockets(self, protein_file: str) -> List[CoordinateBox]:
        """Find list of binding pockets on protein.

        Parameters
        ----------
        protein_file: str
            Protein to load in.

        Returns
        -------
        List[CoordinateBox]
            List of binding pockets on protein. Each pocket is a `CoordinateBox`
        """
        coords, _ = load_molecule(protein_file)
        return get_face_boxes(coords, self.pad)

    def find_pockets(self, macromolecule_file: str) -> List[CoordinateBox]:
        """Find list of suitable binding pockets on protein.

        This function computes putative binding pockets on this protein.
        This class uses the `ConvexHull` to compute binding pockets. Each
        face of the hull is converted into a coordinate box used for
        binding.

        Parameters
        ----------
        macromolecule_file: str
            Location of the macromolecule file to load

        Returns
        -------
        List[CoordinateBox]
            List of pockets. Each pocket is a `CoordinateBox`
        """
        coords, _ = load_molecule(macromolecule_file,
                                  add_hydrogens=False,
                                  calc_charges=False)
        boxes = get_face_boxes(coords, self.pad)
        boxes = merge_overlapping_boxes(boxes)
        return boxes



"""
Normalizing flows for transforming probability distributions.
"""

import logging
from typing import List, Optional, Sequence, Callable

import tensorflow as tf

from deepchem.models.keras_model import KerasModel
from deepchem.utils.typing import OneOrMany
from deepchem.utils.data_utils import load_from_disk, save_to_disk

logger = logging.getLogger(__name__)


class NormalizingFlow(tf.keras.models.Model):
    """Base class for normalizing flow.

    The purpose of a normalizing flow is to map a simple distribution (that is
    easy to sample from and evaluate probability densities for) to a more
    complex distribution that is learned from data. The base distribution
    p(x) is transformed by the associated normalizing flow y=g(x) to model the
    distribution p(y).

    Normalizing flows combine the advantages of autoregressive models
    (which provide likelihood estimation but do not learn features) and
    variational autoencoders (which learn feature representations but
    do not provide marginal likelihoods).

    """

    def __init__(self, base_distribution, flow_layers: Sequence,
                 **kwargs) -> None:
        """Create a new NormalizingFlow.

        Parameters
        ----------
        base_distribution: tfd.Distribution
            Probability distribution to be transformed.
            Typically an N dimensional multivariate Gaussian.
        flow_layers: Sequence[tfb.Bijector]
            An iterable of bijectors that comprise the flow.
        **kwargs

        """

        try:
            import tensorflow_probability as tfp
            tfd = tfp.distributions
            tfb = tfp.bijectors
        except ModuleNotFoundError:
            raise ImportError(
                "This class requires tensorflow-probability to be installed.")

        self.base_distribution = base_distribution
        self.flow_layers = flow_layers

        # Chain of flows is also a normalizing flow
        bijector = tfb.Chain(list(reversed(self.flow_layers)))

        # An instance of tfd.TransformedDistribution
        self.flow = tfd.TransformedDistribution(
            distribution=self.base_distribution, bijector=bijector)

        super(NormalizingFlow, self).__init__(**kwargs)

    def __call__(self, *inputs, training=True):
        return self.flow.bijector.forward(*inputs)


class NormalizingFlowModel(KerasModel):
    """A base distribution and normalizing flow for applying transformations.

    Normalizing flows are effective for any application requiring
    a probabilistic model that can both sample from a distribution and
    compute marginal likelihoods, e.g. generative modeling,
    unsupervised learning, or probabilistic inference. For a thorough review
    of normalizing flows, see [1]_.

    A distribution implements two main operations:
        1. Sampling from the transformed distribution
        2. Calculating log probabilities

    A normalizing flow implements three main operations:
        1. Forward transformation
        2. Inverse transformation
        3. Calculating the Jacobian

    Deep Normalizing Flow models require normalizing flow layers where
    input and output dimensions are the same, the transformation is invertible,
    and the determinant of the Jacobian is efficient to compute and
    differentiable. The determinant of the Jacobian of the transformation
    gives a factor that preserves the probability volume to 1 when transforming
    between probability densities of different random variables.

    References
    ----------
    .. [1] Papamakarios, George et al. "Normalizing Flows for Probabilistic Modeling and Inference." (2019). https://arxiv.org/abs/1912.02762.

    """

    def __init__(self, model: NormalizingFlow, **kwargs) -> None:
        """Creates a new NormalizingFlowModel.

        In addition to the following arguments, this class also accepts all the keyword arguments from KerasModel.

        Parameters
        ----------
        model: NormalizingFlow
            An instance of NormalizingFlow.

        Examples
        --------
        >> import tensorflow_probability as tfp
        >> tfd = tfp.distributions
        >> tfb = tfp.bijectors
        >> flow_layers = [
        ..    tfb.RealNVP(
        ..        num_masked=2,
        ..        shift_and_log_scale_fn=tfb.real_nvp_default_template(
        ..            hidden_layers=[8, 8]))
        ..]
        >> base_distribution = tfd.MultivariateNormalDiag(loc=[0., 0., 0.])
        >> nf = NormalizingFlow(base_distribution, flow_layers)
        >> nfm = NormalizingFlowModel(nf)
        >> dataset = NumpyDataset(
        ..    X=np.random.rand(5, 3).astype(np.float32),
        ..    y=np.random.rand(5,),
        ..    ids=np.arange(5))
        >> nfm.fit(dataset)

        """

        try:
            import tensorflow_probability as tfp
            _ = tfp.distributions
            _ = tfp.bijectors
        except ModuleNotFoundError:
            raise ImportError(
                "This class requires tensorflow-probability to be installed.")

        self.nll_loss_fn = lambda input, labels, weights: self.create_nll(input)

        super(NormalizingFlowModel, self).__init__(model=model,
                                                   loss=self.nll_loss_fn,
                                                   **kwargs)

        self.flow = self.model.flow  # normalizing flow

        # TODO: Incompability between TF and TFP means that TF doesn't track
        # trainable variables in the flow; must override `_create_gradient_fn`
        # self._variables = self.flow.trainable_variables

    def create_nll(self, input: OneOrMany[tf.Tensor]) -> tf.Tensor:
        """Create the negative log likelihood loss function.

        The default implementation is appropriate for most cases. Subclasses can
        override this if there is a need to customize it.

        Parameters
        ----------
        input: OneOrMany[tf.Tensor]
            A batch of data.

        Returns
        -------
        A Tensor equal to the loss function to use for optimization.

        """

        return -tf.reduce_mean(self.flow.log_prob(input, training=True))

    def save(self):
        """Saves model to disk using joblib."""
        save_to_disk(self.model, self.get_model_filename(self.model_dir))

    def reload(self):
        """Loads model from joblib file on disk."""
        self.model = load_from_disk(self.get_model_filename(self.model_dir))

    def _create_gradient_fn(self,
                            variables: Optional[List[tf.Variable]]) -> Callable:
        """Create a function that computes gradients and applies them to the model.

        Because of the way TensorFlow function tracing works, we need to create a
        separate function for each new set of variables.

        Parameters
        ----------
        variables: Optional[List[tf.Variable]]
            Variables to track during training.

        Returns
        -------
        Callable function that applies gradients for batch of training data.

        """

        @tf.function(experimental_relax_shapes=True)
        def apply_gradient_for_batch(inputs, labels, weights, loss):
            with tf.GradientTape() as tape:
                tape.watch(self.flow.trainable_variables)
                if isinstance(inputs, tf.Tensor):
                    inputs = [inputs]
                if self._loss_outputs is not None:
                    inputs = [inputs[i] for i in self._loss_outputs]
                batch_loss = loss(inputs, labels, weights)
            if variables is None:
                vars = self.flow.trainable_variables
            else:
                vars = variables
            grads = tape.gradient(batch_loss, vars)
            self._tf_optimizer.apply_gradients(zip(grads, vars))
            self._global_step.assign_add(1)
            return batch_loss

        return apply_gradient_for_batch


class NormalizingFlowLayer(object):
    """Base class for normalizing flow layers.

    This is an abstract base class for implementing new normalizing flow
    layers that are not available in tfb. It should not be called directly.

    A normalizing flow transforms random variables into new random variables.
    Each learnable layer is a bijection, an invertible
    transformation between two probability distributions. A simple initial
    density is pushed through the normalizing flow to produce a richer,
    more multi-modal distribution. Normalizing flows have three main operations:

    1. Forward
        Transform a distribution. Useful for generating new samples.
    2. Inverse
        Reverse a transformation, useful for computing conditional probabilities.
    3. Log(|det(Jacobian)|) [LDJ]
        Compute the determinant of the Jacobian of the transformation,
        which is a scaling that conserves the probability "volume" to equal 1.

    For examples of customized normalizing flows applied to toy problems,
    see [1]_.

    References
    ----------
    .. [1] Saund, Brad. "Normalizing Flows." (2020). https://github.com/bsaund/normalizing_flows.

    Notes
    -----
    - A sequence of normalizing flows is a normalizing flow.
    - The Jacobian is the matrix of first-order derivatives of the transform.

    """

    def __init__(self, **kwargs):
        """Create a new NormalizingFlowLayer."""

        pass

    def _forward(self, x: tf.Tensor) -> tf.Tensor:
        """Forward transformation.

        x = g(y)

        Parameters
        ----------
        x: tf.Tensor
            Input tensor.

        Returns
        -------
        fwd_x: tf.Tensor
            Transformed tensor.

        """

        raise NotImplementedError("Forward transform must be defined.")

    def _inverse(self, y: tf.Tensor) -> tf.Tensor:
        """Inverse transformation.

        x = g^{-1}(y)

        Parameters
        ----------
        y: tf.Tensor
            Input tensor.

        Returns
        -------
        inv_y: tf.Tensor
            Inverted tensor.

        """

        raise NotImplementedError("Inverse transform must be defined.")

    def _forward_log_det_jacobian(self, x: tf.Tensor) -> tf.Tensor:
        """Log |Determinant(Jacobian(x)|

        Note x = g^{-1}(y)

        Parameters
        ----------
        x: tf.Tensor
            Input tensor.

        Returns
        -------
        ldj: tf.Tensor
            Log of absolute value of determinant of Jacobian of x.

        """

        raise NotImplementedError("LDJ must be defined.")

    def _inverse_log_det_jacobian(self, y: tf.Tensor) -> tf.Tensor:
        """Inverse LDJ.

        The ILDJ = -LDJ.

        Note x = g^{-1}(y)

        Parameters
        ----------
        y: tf.Tensor
            Input tensor.

        Returns
        -------
        ildj: tf.Tensor
            Log of absolute value of determinant of Jacobian of y.

        """

        return -self._forward_log_det_jacobian(self._inverse(y))



"""
Density Functional Theory Data
Derived from: https://github.com/mfkasim1/xcnn/blob/f2cb9777da2961ac553f256ecdcca3e314a538ca/xcdnn2/entry.py
"""
from __future__ import annotations
from abc import abstractmethod, abstractproperty
from typing import List, Dict, Optional
import numpy as np

# dqc depend
import dqc
from dqc.system.mol import Mol
from dqc.system.base_system import BaseSystem
from dqc.grid.base_grid import BaseGrid
from deepchem.utils.dftutils import KSCalc


class DFTSystem():
    """
    The DFTSystem class creates and returns the various systems in an entry object as dictionaries.

    Examples
    --------
    >>> from deepchem.feat.dft_data import DFTSystem
    >>> systems = {'moldesc': 'Li 1.5070 0 0; H -1.5070 0 0','basis': '6-311++G(3df,3pd)'}
    >>> output = DFTSystem(systems)

    Returns
    -------
    DFTSystem object for all the individual atoms/ions/molecules in an entry object.

    References
    ----------
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation functional from nature with fully differentiable density functional theory." Physical Review Letters 127.12 (2021): 126403.

    https://github.com/diffqc/dqc/blob/0fe821fc92cb3457fb14f6dff0c223641c514ddb/dqc/system/base_system.py
    """

    def __init__(self, system: Dict):
        self.system = system
        self.moldesc = system["moldesc"]
        self.basis = system["basis"]
        self.spin = 0
        self.charge = 0
        if 'spin' in system.keys():
            self.spin = int(system["spin"])
        if 'charge' in system.keys():
            self.charge = int(system["charge"])
        """
        Parameters
        ----------
        system: Dict
            system is a dictionary containing information on the atomic positions,
            atomic numbers, and basis set used for the DFT calculation.
        """

    def get_dqc_mol(self, pos_reqgrad: bool = False) -> BaseSystem:
        """
        This method converts the system dictionary to a DQC system and returns it.
        Parameters
        ----------
        pos_reqgrad: bool
            decides if the atomic position require gradient calculation.
        Returns
        -------
        mol
            DQC mol object
        """
        atomzs, atomposs = dqc.parse_moldesc(self.moldesc)
        if pos_reqgrad:
            atomposs.requires_grad_()
        mol = Mol(self.moldesc, self.basis, spin=self.spin, charge=self.charge)
        return mol


class DFTEntry():
    """
    Handles creating and initialising DFTEntry objects from the dataset. This object contains information    about the various systems in the datapoint (atoms, molecules and ions) along with the ground truth
    values.
    Notes
    -----
    Entry class should not be initialized directly, but created through
    ``Entry.create``

    Example
    -------
    >>> from deepchem.feat.dft_data import DFTEntry
    >>> e_type= 'dm'
    >>> true_val= 'deepchem/data/tests/dftHF_output.npy'
    >>> systems = [{'moldesc': 'H 0.86625 0 0; F -0.86625 0 0','basis': '6-311++G(3df,3pd)'}]
    >>> dm_entry_for_HF = DFTEntry.create(e_type, true_val, systems)
    """

    @classmethod
    def create(self, e_type: str, true_val: str, systems: List[Dict]):
        """
        This method is used to initialise the DFTEntry class. The entry objects are created
        based on their entry type.

        Parameters
        ----------
        e_type: str
            Determines the type of calculation to be carried out on the entry
            object. Accepts the following values: "ae", "ie", "dm", "dens", that stand for atomization energy,
            ionization energy, density matrix and density profile respectively.
        true_val: str
            Ground state energy values for the entry object as a string (for ae
            and ie), or a .npy file containing a matrix ( for dm and dens).
        systems: List[Dict]
            List of dictionaries contains "moldesc", "basis" and "spin"
            of all the atoms/molecules. These values are to be entered in
            the DQC or PYSCF format. The systems needs to be entered in a
            specific order, i.e ; the main atom/molecule needs to be the
            first element. (This is for objects containing equations, such
            as ae and ie entry objects). Spin and charge of the system are
            optional parameters and are considered '0' if not specified.
        Returns
        -------
        DFTEntry object based on entry type

        """
        if e_type == "ae":
            return _EntryAE(e_type, true_val, systems)
        elif e_type == "ie":
            return _EntryIE(e_type, true_val, systems)
        elif e_type == "dm":
            return _EntryDM(e_type, true_val, systems)
        elif e_type == "dens":
            return _EntryDens(e_type, true_val, systems)
        else:
            raise NotImplementedError("Unknown entry type: %s" % e_type)

    def __init__(self, e_type: str, true_val: str, systems: List[Dict]):
        self._systems = [DFTSystem(p) for p in systems]

    def get_systems(self) -> List[DFTSystem]:
        """
        Parameters
        ----------
        systems:  List[DFTSystem]

        Returns
        -------
        List of systems in the entry
        """
        return self._systems

    @abstractproperty
    def entry_type(self) -> str:
        """
        Returns
        -------
        The type of entry ;
        1) Atomic Ionization Potential (IP/IE)
        2) Atomization Energy (AE)
        3) Density Profile (DENS)
        4) Density Matrix (DM)
        """
        pass

    def get_true_val(self) -> np.ndarray:
        """
        Get the true value of the DFTEntry.
        For the AE and IP entry types, the experimental values are collected from the NIST CCCBDB/ASD
        databases.
        The true values of density profiles are calculated using PYSCF-CCSD calculations. This method            simply loads the value, no calculation is performed.
        """
        return np.array(0)

    @abstractmethod
    def get_val(self, qcs: List[KSCalc]) -> np.ndarray:
        """
        Return the energy value of the entry, using a DQC-DFT calculation, where the XC has been
        replaced by the trained neural network. This method does not carry out any calculations, it is
        an interface to the KSCalc utility.
        """
        pass


class _EntryDM(DFTEntry):
    """
    Entry for Density Matrix (DM)
    The density matrix is used to express total energy of both non-interacting and
    interacting systems.

    Notes
    -----
    dm entry can only have 1 system
    """

    def __init__(self, e_type, true_val, systems):
        """
        Parameters
        ----------
        e_type: str
        true_val: str
           must be a .npy file containing the pre-calculated density matrix
        systems: List[Dict]

        """
        super().__init__(e_type, true_val, systems)
        self.true_val = true_val
        assert len(self.get_systems()) == 1

    @property
    def entry_type(self) -> str:
        return "dm"

    def get_true_val(self) -> np.ndarray:
        # get the density matrix from PySCF's CCSD calculation
        dm = np.load(self.true_val)
        return dm

    def get_val(self, qcs: List[KSCalc]) -> np.ndarray:
        val = qcs[0].aodmtot()
        return np.array([val.tolist()])


class _EntryDens(DFTEntry):
    """
    Entry for density profile (dens), compared with CCSD calculation
    """

    def __init__(self, e_type, true_val, systems):
        """
        Parameters
        ----------
        e_type: str
        true_val: str
           must be a .npy file containing the pre-calculated density profile.
        systems: List[Dict]
        """
        super().__init__(e_type, true_val, systems)
        self.true_val = true_val
        assert len(self.get_systems()) == 1
        self._grid: Optional[BaseGrid] = None

    @property
    def entry_type(self) -> str:
        return "dens"

    def get_true_val(self) -> np.ndarray:
        dens = np.load(self.true_val)
        return dens

    def get_val(self, qcs: List[KSCalc]) -> np.ndarray:
        """
        This method calculates the integration grid which is then used to calculate the
        density profile of an entry object.

        Parameters
        ----------
        qcs: List[KSCalc]

        Returns
        -------
        Numpy array containing calculated density profile
        """
        qc = qcs[0]

        grid = self.get_integration_grid()
        rgrid = grid.get_rgrid()

        return (qc.dens(rgrid)).numpy()

    def get_integration_grid(self) -> BaseGrid:
        """
        This method is used to calculate the integration grid required for a system
        in order to calculate it's density profile using Differentiable DFT.

        Returns
        -------
        grid: BaseGrid

        References
        ----------
        https://github.com/diffqc/dqc/blob/0fe821fc92cb3457fb14f6dff0c223641c514ddb/dqc/grid/base_grid.py
        """
        if self._grid is None:
            system = self.get_systems()[0]
            dqc_mol = system.get_dqc_mol()
            dqc_mol.setup_grid()
            grid = dqc_mol.get_grid()
            assert grid.coord_type == "cart"
            self._grid = grid

        return self._grid


class _EntryIE(DFTEntry):
    """
    Entry for Ionization Energy (IE)
    """

    def __init__(self, e_type, true_val, systems):
        """
        Parameters
        ----------
        e_type: str
        true_val: str
        systems: List[Dict]
        """
        super().__init__(e_type, true_val, systems)
        self.true_val = float(true_val)

    @property
    def entry_type(self) -> str:
        return "ie"

    def get_true_val(self) -> np.ndarray:
        return np.array([self.true_val])

    def get_val(self, qcs: List[KSCalc]) -> np.ndarray:
        """
        This method calculates the energy of an entry based on the systems and command present
        in the data object. For example; for a Lithium hydride molecule the total energy
        would be ; energy(Li) + energy(H) - energy(LiH)

        Parameters
        ----------
        qcs: List[KSCalc]

        Returns
        -------
        Total Energy of a data object for entry types IE and AE
        """
        e = [m.energy() for m in qcs]
        val = sum(e) - 2 * e[0]
        return np.array([val.tolist()])


class _EntryAE(_EntryIE):
    """
    Entry for Atomization Energy (AE)
    """

    @property
    def entry_type(self) -> str:
        return "ae"
